#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_trading.constants as constants
import octobot_trading.personal_data.portfolios.assets.margin_asset as margin_asset
import octobot_trading.personal_data.portfolios.portfolio as portfolio_class


class MarginPortfolio(portfolio_class.Portfolio):
    def create_currency_asset(self, currency, available=constants.ZERO, total=constants.ZERO):
        return margin_asset.MarginAsset(name=currency, available=available, total=total)

    def update_portfolio_data_from_order(self, order):
        pass  # TODO

    def update_portfolio_data_from_withdrawal(self, amount, currency):
        pass  # TODO

    def update_portfolio_available_from_order(self, order, is_new_order=True):
        pass  # TODO



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import octobot_commons.logging as logging

import octobot_trading.util as util


class MarketsManager(util.Initializable):
    def __init__(self):  # Required for python development
        super().__init__()
        self.logger: logging.BotLogger = logging.get_logger(self.__class__.__name__)
        self.markets: dict[str, any] = {}

    async def initialize_impl(self):
        self._reset_kline()

    def _reset_kline(self):
        self.markets = {}

    def markets_update(self, markets):
        self.markets = markets



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest

import octobot_commons.enums as common_enums
from octobot_trading.enums import ExchangeConstantsMarketStatusColumns as Ecmsc, \
    ExchangeConstantsOrderBookInfoColumns as Ecobic, ExchangeConstantsOrderColumns as Ecoc, \
    ExchangeConstantsTickersColumns as Ectc
from tests_additional.real_exchanges.real_option_exchange_tester import RealOptionExchangeTester
import octobot_trading.exchanges.connectors.ccxt.enums as ccxt_enums

# required to catch async loop context exceptions
from tests import event_loop

try:
    from tentacles.Trading.Exchange.polymarket.ccxt.polymarket_async import polymarket
except ImportError:
    # test will be skipped if the tentacle is not installed
    pytest.skip(
        "Polymarket tentacle is not installed, skipping TestPolymarketRealExchangeTester",
        allow_module_level=True
    )

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class TestPolymarketRealExchangeTester(RealOptionExchangeTester):
    EXCHANGE_NAME = "polymarket"
    SYMBOL = f"will-bitcoin-replace-sha-256-before-2027/USDC:USDC-261231-0-YES"
    SYMBOL_2 = f"will-the-us-confirm-that-aliens-exist-before-2027/USDC:USDC-261231-0-YES"
    SYMBOL_3 = f"10pt0-or-above-earthquake-before-2027/USDC:USDC-261231-0-YES"
    TIME_FRAME = common_enums.TimeFrames.ONE_MINUTE
    USES_TENTACLE = True  # set True when an exchange tentacles should be used in this test
    PROFILE_ID = "0x16b29c50f2439faf627209b2ac0c7bbddaa8a881" # https://polymarket.com/@SeriouslySirius

    async def test_time_frames(self):
        time_frames = await self.time_frames()
        assert all(time_frame in time_frames for time_frame in (
            common_enums.TimeFrames.ONE_MINUTE.value,
            common_enums.TimeFrames.ONE_HOUR.value,
            common_enums.TimeFrames.SIX_HOURS.value,
            common_enums.TimeFrames.ONE_DAY.value,
            common_enums.TimeFrames.ONE_WEEK.value,
        ))

    async def test_active_symbols(self):
        await self.inner_test_active_symbols(17000, 19000)

    async def test_get_market_status(self):
        # Debug code to print the symbol of the slugs
        # slugs = ['10pt0-or-above-earthquake-before-2027', 'will-bitcoin-replace-sha-256-before-2027', 'will-the-us-confirm-that-aliens-exist-before-2027']
        # async with self.get_exchange_manager() as exchange_manager:
        #     for market in exchange_manager.exchange.connector.client.markets:
        #         for slug in slugs:
        #             if market.startswith(slug):
        #                 print(exchange_manager.exchange.connector.client.markets[market])
        for market_status in await self.get_market_statuses():
            self.ensure_required_market_status_values(market_status)
            self.check_market_status_limits(market_status, has_price_limits=True)

    async def test_get_symbol_prices(self):
        # without limit
        symbol_prices = await self.get_symbol_prices()
        assert len(symbol_prices) >= 1000
        # check candles order (oldest first)
        self.ensure_elements_order(symbol_prices, common_enums.PriceIndexes.IND_PRICE_TIME.value)
        # check last candle is the current candle
        assert symbol_prices[-1][common_enums.PriceIndexes.IND_PRICE_TIME.value] >= self.get_time() - self.get_allowed_time_delta()

        # try with candles limit (used in candled updater)
        symbol_prices = await self.get_symbol_prices(limit=200)
        assert len(symbol_prices) == 200
        # check candles order (oldest first)
        self.ensure_elements_order(symbol_prices, common_enums.PriceIndexes.IND_PRICE_TIME.value)
        # check last candle is the current candle
        assert symbol_prices[-1][common_enums.PriceIndexes.IND_PRICE_TIME.value] >= self.get_time() - self.get_allowed_time_delta()

    async def test_get_historical_symbol_prices(self):
        # try with since and limit (used in data collector)
        for limit in (50, None):
            symbol_prices = await self.get_symbol_prices(since=self.CANDLE_SINCE, limit=limit)
            if limit:
                assert len(symbol_prices) == limit
            else:
                assert len(symbol_prices) > 5
            # check candles order (oldest first)
            self.ensure_elements_order(symbol_prices, common_enums.PriceIndexes.IND_PRICE_TIME.value)
            # check that fetched candles are historical candles
            max_candle_time = self.get_time_after_time_frames(self.CANDLE_SINCE_SEC, len(symbol_prices))
            assert max_candle_time <= self.get_time()
            with pytest.raises(AssertionError):  # not supported
                for candle in symbol_prices:
                    assert self.CANDLE_SINCE_SEC <= candle[common_enums.PriceIndexes.IND_PRICE_TIME.value] <= max_candle_time

    async def test_get_historical_ohlcv(self):
        await super().test_get_historical_ohlcv()

    async def test_get_kline_price(self):
        kline_price = await self.get_kline_price()
        assert len(kline_price) == 1
        assert len(kline_price[0]) == 6
        kline_start_time = kline_price[0][common_enums.PriceIndexes.IND_PRICE_TIME.value]
        # assert kline is the current candle
        assert kline_start_time >= self.get_time() - self.get_allowed_time_delta()

    async def test_get_order_book(self):
        order_book = await self.get_order_book()
        assert 0 < order_book[Ecobic.TIMESTAMP.value] < self._get_ref_order_book_timestamp()
        assert len(order_book[Ecobic.ASKS.value]) >= 5
        assert len(order_book[Ecobic.ASKS.value][0]) == 2
        assert len(order_book[Ecobic.BIDS.value]) >= 5
        assert len(order_book[Ecobic.BIDS.value][0]) == 2
        
    async def test_get_order_books(self):
        await self.inner_test_get_order_books(
            True,
            1000, # asked symbols
            100, # up to 100 orders
            0, # from 0 orders
            False,
            None,
            0,
            False
        )

    async def test_get_recent_trades(self):
        recent_trades = await self.get_recent_trades()
        assert len(recent_trades) == 50
        # check trades order (oldest first)
        self.ensure_elements_order(recent_trades, Ecoc.TIMESTAMP.value)

    async def test_get_price_ticker(self):
        ticker = await self.get_price_ticker()
        self._check_ticker(ticker, self.SYMBOL, check_content=True)

    async def test_get_all_currencies_price_ticker(self):
        tickers = await self.get_all_currencies_price_ticker()
        for symbol, ticker in tickers.items():
            self._check_ticker(ticker, symbol)

    @staticmethod
    def _check_ticker(ticker, symbol, check_content=False):
        assert ticker[Ectc.SYMBOL.value] == symbol
        assert all(key in ticker for key in (
            Ectc.HIGH.value,
            Ectc.LOW.value,
            Ectc.BID.value,
            Ectc.BID_VOLUME.value,
            Ectc.ASK.value,
            Ectc.ASK_VOLUME.value,
            Ectc.OPEN.value,
            Ectc.CLOSE.value,
            Ectc.LAST.value,
            Ectc.PREVIOUS_CLOSE.value
        ))
        if check_content:
            assert ticker[Ectc.HIGH.value] is None
            assert ticker[Ectc.LOW.value] is None
            assert ticker[Ectc.BID.value]
            assert ticker[Ectc.BID_VOLUME.value] is None
            assert ticker[Ectc.ASK.value]
            assert ticker[Ectc.ASK_VOLUME.value] is None
            assert ticker[Ectc.OPEN.value] is None
            assert ticker[Ectc.CLOSE.value]
            assert ticker[Ectc.LAST.value]
            assert ticker[Ectc.PREVIOUS_CLOSE.value] is None
            assert ticker[Ectc.BASE_VOLUME.value]
            assert ticker[Ectc.TIMESTAMP.value]
            RealExchangeTester.check_ticker_typing(
                ticker, 
                check_open=False, 
                check_high=False, 
                check_low=False, 
                check_close=True,
                check_base_volume=True, 
                check_timestamp=True
            )

    async def test_fetch_user_positions(self, **kwargs):
        positions = await self.get_user_positions(**kwargs)
        for position in positions:
            self._check_position(position)

    async def test_fetch_user_closed_positions(self, **kwargs):
        # positions = await self.get_user_closed_positions(**kwargs)
        # for position in positions:
        #     self._check_position(position, is_closed=True)
        pass # Not possible while market statuses cannot be dynamically updated

    def _check_position(self, position, check_symbol=False, is_closed=False):
        assert position
        if check_symbol:
            assert position[ccxt_enums.ExchangePositionCCXTColumns.SYMBOL.value] == self.SYMBOL
        else:
            assert position[ccxt_enums.ExchangePositionCCXTColumns.SYMBOL.value] is not None
        assert position[ccxt_enums.ExchangePositionCCXTColumns.ENTRY_PRICE.value] > 0
        if is_closed:
            assert position[ccxt_enums.ExchangePositionCCXTColumns.TIMESTAMP.value] > 0
        else:
            assert position[ccxt_enums.ExchangePositionCCXTColumns.MARK_PRICE.value] is not None # can be 0.0 on Polymarket
            assert position[ccxt_enums.ExchangePositionCCXTColumns.UNREALISED_PNL.value] is not None
            assert position[ccxt_enums.ExchangePositionCCXTColumns.COLLATERAL.value] >= 0
            assert position[ccxt_enums.ExchangePositionCCXTColumns.NOTIONAL.value] is not None # can be 0.0
            assert position[ccxt_enums.ExchangePositionCCXTColumns.SIDE.value] is not None
        assert position[ccxt_enums.ExchangePositionCCXTColumns.REALISED_PNL.value] is not None # can be 0.0
        assert ccxt_enums.ExchangePositionCCXTColumns.CONTRACT_SIZE.value in position


#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.

from octobot_trading.personal_data.portfolios.channel import balance
from octobot_trading.personal_data.portfolios.channel.balance import (
    BalanceProducer,
    BalanceChannel,
    BalanceProfitabilityProducer,
    BalanceProfitabilityChannel,
)
from octobot_trading.personal_data.portfolios.channel import balance_updater
from octobot_trading.personal_data.portfolios.channel.balance_updater import (
    BalanceUpdater,
    BalanceProfitabilityUpdater,
)
from octobot_trading.personal_data.portfolios.channel import balance_updater_simulator
from octobot_trading.personal_data.portfolios.channel.balance_updater_simulator import (
    BalanceUpdaterSimulator,
    BalanceProfitabilityUpdaterSimulator,
)

__all__ = [
    "BalanceUpdaterSimulator",
    "BalanceProfitabilityUpdaterSimulator",
    "BalanceUpdater",
    "BalanceProfitabilityUpdater",
    "BalanceProducer",
    "BalanceChannel",
    "BalanceProfitabilityProducer",
    "BalanceProfitabilityChannel",
]



#  Drakkar-Software OctoBot-Trading
#  Copyright (c) Drakkar-Software, All rights reserved.
#
#  This library is free software; you can redistribute it and/or
#  modify it under the terms of the GNU Lesser General Public
#  License as published by the Free Software Foundation; either
#  version 3.0 of the License, or (at your option) any later version.
#
#  This library is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#  Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public
#  License along with this library.
import pytest
import mock

from tests import event_loop
from octobot_commons.tests.test_config import load_test_config
from octobot_commons.constants import CONFIG_CRYPTO_CURRENCIES
from octobot_commons.enums import TimeFrames
from octobot_trading.exchanges.exchange_manager import ExchangeManager
from octobot_trading.api.exchange import cancel_ccxt_throttle_task
import octobot_trading.exchange_channel as exchange_channel
import octobot_trading.constants as trading_constants

pytestmark = pytest.mark.asyncio


class TestExchangeConfig:
    EXCHANGE_NAME = "binanceus"

    @staticmethod
    async def init_default(config=None):
        if not config:
            config = load_test_config()

        exchange_manager = ExchangeManager(config, TestExchangeConfig.EXCHANGE_NAME)

        await exchange_manager.initialize(exchange_config_by_exchange=None)
        return config, exchange_manager

    async def test_traded_pairs(self):
        config = load_test_config()
        config[CONFIG_CRYPTO_CURRENCIES] = {
            "Avalanche": {
                "pairs": ["AVAX/BTC"]
            },
            "Ethereum": {
                "enabled": True,
                "pairs": ["ETH/USDT"]
            },
            "Uniswap": {
                "enabled": False,
                "pairs": ["UNI/BTC"]
            }
        }

        _, exchange_manager = await self.init_default(config=config)
        try:
            assert exchange_manager.exchange_config.traded_cryptocurrencies == {
                "Ethereum": ["ETH/USDT"],
                "Avalanche": ["AVAX/BTC"]
            }
            all_pairs = sorted(["AVAX/BTC", "ETH/USDT", "UNI/BTC"])
            all_enabled_pairs = sorted(["AVAX/BTC", "ETH/USDT"])
            assert sorted(exchange_manager.exchange_config.traded_symbol_pairs) == all_enabled_pairs
            cancel_ccxt_throttle_task()
        finally:
            await exchange_manager.stop()

    async def test_traded_pairs_with_wildcard(self):
        config = load_test_config()
        config[CONFIG_CRYPTO_CURRENCIES] = {
            "Bitcoin": {
                "pairs": ["*"],
                "quote": "BTC"
            },
            "Ethereum": {
                "enabled": False,
                "pairs": ["*"],
                "quote": "ETH"
            }
        }
        _, exchange_manager = await self.init_default(config=config)
        try:
            assert "AVAX/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "ADA/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "MATIC/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "ONT/BTC" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "BTC/USDT" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "ETH/USDT" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "AVAX/BNB" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "ETH/BTC" in exchange_manager.exchange_config.traded_symbol_pairs

            # inactive markets
            assert "UNI/BTC" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]

            # disabled
            assert "Ethereum" not in exchange_manager.exchange_config.traded_cryptocurrencies
            assert "ADA/ETH" not in exchange_manager.exchange_config.traded_symbol_pairs

            cancel_ccxt_throttle_task()
        finally:
            await exchange_manager.stop()

    async def test_traded_pairs_with_invalid_wildcard(self):
        config = load_test_config()

        # missing quote key
        config[CONFIG_CRYPTO_CURRENCIES] = {
            "Bitcoin": {
                "enabled": True,
                "pairs": ["*"],
                "quote": "BTC"
            },
            "Ethereum": {
                "pairs": ["*"],
            }
        }
        _, exchange_manager = await self.init_default(config=config)
        try:
            assert "ADA/BTC" in exchange_manager.exchange_config.traded_symbol_pairs
            assert "Bitcoin" in exchange_manager.exchange_config.traded_cryptocurrencies

            # inactive markets
            assert "TRX/BTC" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]

            # invalid ETH wildcard config
            assert "Ethereum" not in exchange_manager.exchange_config.traded_cryptocurrencies

            cancel_ccxt_throttle_task()
        finally:
            await exchange_manager.stop()

    async def test_traded_pairs_with_add(self):
        config = load_test_config()
        config[CONFIG_CRYPTO_CURRENCIES] = {
            "Bitcoin": {
                "pairs": ["*"],
                "quote": "BTC",
                "add": ["BTC/USDT"]
            },
            "Ethereum": {
                "enabled": False,
                "pairs": ["*"],
                "quote": "ETH",
                "add": ["ETH/USDT"]
            }
        }

        _, exchange_manager = await self.init_default(config=config)
        try:
            assert "AVAX/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "ADA/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "MATIC/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "LINK/BTC" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "ONT/BTC" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "BTC/USDT" in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "ETH/USDT" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "AVAX/BNB" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]
            assert "BTC/USDT" in exchange_manager.exchange_config.traded_symbol_pairs

            # inactive markets
            assert "UNI/BTC" not in exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"]

            # disabled
            assert "Ethereum" not in exchange_manager.exchange_config.traded_cryptocurrencies
            assert "ADA/ETH" not in exchange_manager.exchange_config.traded_symbol_pairs
            assert "ETH/USDT" not in exchange_manager.exchange_config.traded_symbol_pairs
            cancel_ccxt_throttle_task()
        finally:
            await exchange_manager.stop()

    async def test_traded_pairs_with_redundancy(self):
        config = load_test_config()
        config[CONFIG_CRYPTO_CURRENCIES] = {
            "Binance Coin": {
                "pairs": [
                    "BNB/USDT"
                ]
            },
            "Binance USD": {
                "pairs": [
                    "BNB/BUSD"
                ]
            },
            "Bitcoin": {
                "enabled": True,
                "pairs": [
                    "BNB/BTC"
                ]
            },
            "Tether": {
                "enabled": True,
                "pairs": [
                    "BNB/USDT"
                ]
            }
        }

        _, exchange_manager = await self.init_default(config=config)
        try:
            assert exchange_manager.exchange_config.traded_cryptocurrencies["Binance Coin"] == ["BNB/USDT"]
            assert exchange_manager.exchange_config.traded_cryptocurrencies["Bitcoin"] == ["BNB/BTC"]
            assert exchange_manager.exchange_config.traded_cryptocurrencies["Tether"] == ["BNB/USDT"]

            # inactive markets
            assert exchange_manager.exchange_config.traded_cryptocurrencies["Binance USD"] == []

            sorted_pairs_without_redundancy = sorted(["BNB/USDT", "BNB/BTC"])
            assert sorted(exchange_manager.exchange_config.traded_symbol_pairs) == sorted_pairs_without_redundancy

            cancel_ccxt_throttle_task()
        finally:
            await exchange_manager.stop()

    @pytest.mark.parametrize("watch_only,added_pairs,removed_pairs,added_time_frames", [
        (False, ["ADA/USDT"], [], [TimeFrames.ONE_MINUTE, TimeFrames.THREE_MINUTES]),
        (True, ["ETH/USDT"], [], [TimeFrames.ONE_MINUTE, TimeFrames.THREE_MINUTES]),
        (False, ["ADA/USDT", "LINK/USDT"], [], []),
        (False, [], ["BTC/USDT"], [TimeFrames.ONE_MINUTE]),
        (True, [], ["BTC/USDT"], [TimeFrames.ONE_MINUTE]),
    ])
    async def test_update_traded_symbol_pairs(self, watch_only, added_pairs, removed_pairs, added_time_frames):
        expected_watch_only_modified_channels = [trading_constants.TICKER_CHANNEL]
        expected_traded_modified_channels = [
            trading_constants.TICKER_CHANNEL,
            trading_constants.OHLCV_CHANNEL,
            trading_constants.FUNDING_CHANNEL,
        ]
        config = load_test_config()
        config[CONFIG_CRYPTO_CURRENCIES] = {
            "Bitcoin": {
                "pairs": ["BTC/USDT"]
            }
        }
        _, exchange_manager = await self.init_default(config=config)
        try:
            assert exchange_manager.exchange_config.additional_time_frames == []
            # init additional_time_frames if a random time frame that is not in available time frames
            assert TimeFrames.ONE_YEAR not in exchange_manager.exchange_config.available_time_frames
            exchange_manager.exchange_config.additional_time_frames = [TimeFrames.ONE_YEAR]
            initial_available_time_frames = list(exchange_manager.exchange_config.available_time_frames)
            assert TimeFrames.ONE_MINUTE in initial_available_time_frames # 1m is already in available time frames
            assert TimeFrames.THREE_MINUTES not in initial_available_time_frames # 3m is NOT in available time frames
            new_added_time_frames = [tf for tf in added_time_frames if tf not in initial_available_time_frames]
            
            # Pre-populate lists for removal tests
            if removed_pairs:
                for pair in removed_pairs:
                    if pair not in exchange_manager.exchange_config.traded_symbol_pairs:
                        exchange_manager.exchange_config.traded_symbol_pairs.append(pair)
                if watch_only:
                    exchange_manager.exchange_config.additional_traded_pairs = removed_pairs.copy()
                    exchange_manager.exchange_config.watched_pairs = removed_pairs.copy()
                else:
                    exchange_manager.exchange_config.additional_traded_pairs = removed_pairs.copy()
            for channel in expected_traded_modified_channels:
                channel_instance = exchange_channel.get_chan(channel, exchange_manager.id)
                channel_instance.modify = mock.AsyncMock(wraps=channel_instance.modify)
            ohlcv_producer = exchange_channel.get_chan(
                trading_constants.OHLCV_CHANNEL, exchange_manager.id
            ).producers[0]
            ticker_producer = exchange_channel.get_chan(
                trading_constants.TICKER_CHANNEL, exchange_manager.id
            ).producers[0]
            with mock.patch.object(
                exchange_manager.exchange_config, '_is_valid_symbol', wraps=exchange_manager.exchange_config._is_valid_symbol
            ) as is_valid_mock, mock.patch.object(
                exchange_manager, 'time_frame_exists', wraps=exchange_manager.time_frame_exists
            ) as time_frame_exists_mock, mock.patch.object(
                ohlcv_producer, '_initialize', mock.AsyncMock()
            ) as ohlcv_updater_initialize_mock, mock.patch.object(
                ticker_producer, 'fetch_and_push_pair', mock.AsyncMock()
            ) as ticker_fetch_and_push_pair_mock, mock.patch.object(
                exchange_manager.exchange_config, '_logger', wraps=exchange_manager.exchange_config._logger
            ) as logger_mock:
                await exchange_manager.exchange_config.update_traded_symbol_pairs(
                    added_pairs=added_pairs,
                    removed_pairs=removed_pairs,
                    added_time_frames=added_time_frames,
                    watch_only=watch_only,
                )
                
                # Verify _is_valid_symbol was called for all pairs
                assert is_valid_mock.call_count == len(added_pairs) + len(removed_pairs)
                assert time_frame_exists_mock.call_count == len(new_added_time_frames)
                if added_pairs:
                    # ticker update called for each added pair, regardless of watch_only or not
                    assert sorted(
                        [call.args[0] for call in ticker_fetch_and_push_pair_mock.mock_calls]
                    ) == sorted(added_pairs)
                if watch_only:
                    ohlcv_updater_initialize_mock.assert_not_called()
                elif added_pairs:
                    # called once for config pairs, once for additional pairs
                    assert len(ohlcv_updater_initialize_mock.mock_calls) == 1
                    assert ohlcv_updater_initialize_mock.mock_calls[0].args == (
                        [pair for pair in added_pairs if pair not in exchange_manager.exchange_config.traded_symbol_pairs],
                        False
                    )
            
            if watch_only:
                # watched pairs should not have any additional time frames, even if provided
                assert exchange_manager.exchange_config.additional_time_frames == [TimeFrames.ONE_YEAR]
                for pair in added_pairs:
                    assert pair in exchange_manager.exchange_config.watched_pairs
                for pair in removed_pairs:
                    assert pair not in exchange_manager.exchange_config.watched_pairs
                self._assert_modified_channels(
                    exchange_manager.exchange_config, expected_watch_only_modified_channels, expected_traded_modified_channels
                )
            else:
                # additional traded pairs should have additional time frames if provided
                if added_time_frames and added_pairs:
                    assert exchange_manager.exchange_config.additional_time_frames == [TimeFrames.ONE_YEAR] + [
                        tf for tf in added_time_frames if tf not in initial_available_time_frames
                    ]
                else:
                    assert exchange_manager.exchange_config.additional_time_frames == [TimeFrames.ONE_YEAR]
                # initial available time frames should not have changed, only additional time frames can change
                assert exchange_manager.exchange_config.available_time_frames == initial_available_time_frames
                for pair in added_pairs:
                    assert pair in exchange_manager.exchange_config.additional_traded_pairs
                for pair in removed_pairs:
                    assert pair not in exchange_manager.exchange_config.additional_traded_pairs
                self._assert_modified_channels(
                    exchange_manager.exchange_config, expected_traded_modified_channels, expected_watch_only_modified_channels
                )

            # ensure no hidden error occured (due to the global error catching)
            if new_added_time_frames and watch_only:
                assert logger_mock.error.call_count == 1
                assert f"Ignored additional time frames for watched pairs" in logger_mock.error.call_args[0][0]
            else:
                assert logger_mock.error.call_count == 0
            
            cancel_ccxt_throttle_task()
        finally:
            await exchange_manager.stop()
        
    async def test_update_traded_symbol_pairs_no_available_time_frames(self):
        config = load_test_config()
        config[CONFIG_CRYPTO_CURRENCIES] = {
            "Bitcoin": {
                "pairs": ["BTC/USDT"]
            }
        }
        _, exchange_manager = await self.init_default(config=config)
        try:
            exchange_manager.exchange_config.available_time_frames = []
            assert exchange_manager.exchange_config.additional_time_frames == []
            with mock.patch.object(
                exchange_manager.exchange_config, '_is_valid_symbol', wraps=exchange_manager.exchange_config._is_valid_symbol
            ) as is_valid_mock, mock.patch.object(
                exchange_manager, 'time_frame_exists', wraps=exchange_manager.time_frame_exists
            ) as time_frame_exists_mock, mock.patch.object(
                exchange_channel.get_chan(
                    trading_constants.OHLCV_CHANNEL, exchange_manager.id
                ).producers[0], '_initialize', mock.AsyncMock()
            ) as ohlcv_updater_initialize_mock, mock.patch.object(
                exchange_channel.get_chan(
                    trading_constants.TICKER_CHANNEL, exchange_manager.id
                ).producers[0], 'fetch_and_push_pair', mock.AsyncMock()
            ) as ticker_fetch_and_push_pair_mock, mock.patch.object(
                exchange_manager.exchange_config, '_logger', wraps=exchange_manager.exchange_config._logger
            ) as logger_mock:
                await exchange_manager.exchange_config.update_traded_symbol_pairs(
                    added_pairs=["ETH/USDT"],
                    removed_pairs=[],
                    added_time_frames=[TimeFrames.ONE_HOUR],
                    watch_only=False,
                )
                assert is_valid_mock.call_count == 1
                assert time_frame_exists_mock.call_count == 1
                ohlcv_updater_initialize_mock.assert_called_once_with(["ETH/USDT"], False)
                ticker_fetch_and_push_pair_mock.assert_called_once_with("ETH/USDT")
                assert logger_mock.error.call_count == 0
                assert exchange_manager.exchange_config.additional_time_frames == [TimeFrames.ONE_HOUR]
        finally:
            await exchange_manager.stop()

    def _assert_modified_channels(
        self, exchange_config, expected_modified_channels, expected_unmodified_channels
    ):
        unmodified_channels = [
            chan for chan in expected_unmodified_channels if chan not in expected_modified_channels
        ]
        for channel in expected_modified_channels:
            channel_instance = exchange_channel.get_chan(channel, exchange_config.exchange_manager.id)
            assert channel_instance.modify.call_count > 0, f"{channel} should have been modified"
        for channel in unmodified_channels:
            channel_instance = exchange_channel.get_chan(channel, exchange_config.exchange_manager.id)
            assert channel_instance.modify.call_count == 0, f"{channel} should not have been modified"
